/* @file TCatEvalGainShiftProcessor.cc
 * @brief evaluation of gain shift and obtain correction parameters
 *
 * @date Create        : 2020-07-27 14:08:51 JST
 *       Last Modified : 2020-07-30 18:01:26 JST (ota)
 * @author: Shinsuke OTA <ota@cns.s.u-tokyo.ac.jp>
 */


#include "TCatEvalGainShiftProcessor.h"
#include "TChain.h"
#include "TMath.h"
#include "TArtLocalUtil.h"
#include "TH2F.h"
#include "TEventHeader.h"
#include "TCatTpcTrack.h"
#include "TFile.h"
#include "TROOT.h"
#include "TCatReadoutPadArray.h"
#include "TCatReadoutPad.h"
#include "TH2Poly.h"

ClassImp(art::TCatEvalGainShiftProcessor)

using art::TCatEvalGainShiftProcessor;

#ifdef USE_MPI
namespace art {
   namespace mpi {
      void GetMPISizeRankIfInitialized(int &useMPI, int &npe, int &myrank) {
         useMPI = 0;
         npe = 0;
         myrank = 0;
         MPI_Initialized(&useMPI);
         if (useMPI) {
            MPI_Comm_size(MPI_COMM_WORLD, &npe);
            MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
         }
      }
   }
};
#endif         


TCatEvalGainShiftProcessor::TCatEvalGainShiftProcessor()
{
   Register(fInputFilename("InputFile","name of input tree file",""));
   Register(fOutputFilename("OutputFile","name of output file",""));
   Register(fEventsModulus("EventsModulus","modulus of the number of events",1000.));
   Register(fEventHeaderName("EventHeader","name of event header","eventheader"));
}

TCatEvalGainShiftProcessor::~TCatEvalGainShiftProcessor()
{
}


void TCatEvalGainShiftProcessor::EndOfRun()
{
   // input root file
   const TString& inputFilename = fInputFilename;
   const TString& outputFilename = fOutputFilename;
   if (outputFilename.IsNull()) return;
   TString filename = outputFilename;
   
#ifdef USE_MPI   
   int myrank, npe,useMPI;
   mpi::GetMPISizeRankIfInitialized(useMPI,npe,myrank);
   if (useMPI) filename = Form("%s%d",filename.Data(),myrank);
#endif   

#ifdef USE_MPI
   if (useMPI) MPI_Barrier(MPI_COMM_WORLD);
#endif


   if (TFile *file = (TFile*)gROOT->GetListOfFiles()->FindObject(inputFilename)) {
      file->Close();
   }
   
   // prepare for chain
   TChain *chain = new TChain("tree");
   TClonesArray *track = NULL;
   TEventHeader *header = NULL;
   chain->Add(Form("%s*",inputFilename.Data()));
   chain->SetBranchAddress("atcluster",&track);
   chain->SetBranchAddress((TString)fEventHeaderName,&header);
   chain->GetEntry(0);
   chain->SetBranchStatus("*",0);
   chain->SetBranchStatus("atcluster",1);
   chain->SetBranchStatus((TString)fEventHeaderName,1);

   TTree *tree = chain->GetTree();
   TCatReadoutPadArray* readoutPad = static_cast<TCatReadoutPadArray*>(tree->GetUserInfo()->FindObject("readoutPad"));

   TFile *file = TFile::Open(outputFilename,"recreate");
   TAnalysisInfo::AddTo(file);
   
   // prepare for histogram
   const int nEvents = chain->GetEntriesFast();
   int nBins = TMath::Max(1,TMath::FloorNint(nEvents/fEventsModulus));
   const int nPads = 416;
   std::vector<TH2*> fHists(nPads);
   TString idfmt = Util::ZeroFilledDigitForm(nPads);
   TString hnamefmt = Form("h%s",idfmt.Data());
   TString dnamefmt = Form("d%s",idfmt.Data());
   
   TString htitlefmt = Form("charge vs evtnum (%s)",idfmt.Data ());

   // prepare for histogram
   TDirectory *topdir = file->mkdir(GetName());
   TDirectory *histdir = topdir->mkdir("TimeEvolution");
   TDirectory *corrdir = topdir->mkdir("Corrected");
   TDirectory *diagdir = topdir->mkdir("Diagnosis");
   
   histdir->cd();
   for (int i = 0; i < nPads; ++i) {
      const TString hname = Form(hnamefmt,i);
      const TString htitle = Form(htitlefmt,i);
      fHists[i] = new TH2F(hname,htitle,nBins,0.,nEvents,100,0.,10000.);
   }

   // loop over pads
   for (int ievt = 0; ievt < nEvents; ++ievt) {
      chain->GetEntry(ievt);
      if (!track || !header) continue;
      const int evtid = header->GetEventNumber();
      const int nTracks = track->GetEntriesFast();
      for (int itrack = 0; itrack < nTracks; ++itrack) {
         TCatTpcTrack *aTrack = static_cast<TCatTpcTrack*>(track->UncheckedAt(itrack));
         std::vector<TCatTrackResultPulse*> hits = aTrack->GetHits();
         const int nHits = hits.size();
         if (nHits > 200) continue;
         for (int ihit = 0; ihit < nHits; ++ihit) {
            fHists[hits[ihit]->GetID()]->Fill(evtid, hits[ihit]->GetCharge());
         }
      }
   }

   topdir->cd();
   TH2* hDiag = new TH2F("hDiag","shift vs id (diag)",nPads,-0.5,nPads-0.5,100,0.,1.);
   TH2* hEval = new TH2F("hEval","shift vs id",nPads,-0.5,nPads-0.5,100,0.,2.);
   TH2* hCorrect = new TH2F("hCorrect","shift vs id (corrected)",nPads,-0.5,nPads-0.5,100,0.,1.);
   TH2Poly *hid  = readoutPad->CreateHist("hIDs","id",-50.,50.,-50.,50.,1);
   TH2Poly *hMean = readoutPad->CreateHist("hMean","average charge",-50.,50.,-50.,50.);
   TH2Poly *hHits = readoutPad->CreateHist("hHits","count",-50.,50.,-50.,50.);
   TH2Poly *hMaybeBad = readoutPad->CreateHist("hMaybeBad","Candidate bad pads",-50.,50.,-50.,50.);
   
   topdir->Add(hid);
   topdir->Add(hHits);
   topdir->Add(hMean);
   topdir->Add(hMaybeBad);
   
   
   
   // list edges vs event number 
   corrdir->cd();
   std::vector<TH2*> scaled(nPads);



   // diagnosis using probability 
   for (int i =0;  i < nPads; ++i) {
      std::vector<double> scales;
      std::vector<double> curvature;
      hHits->SetBinContent(i+1,fHists[i]->GetEntries());
      hMean->SetBinContent(i+1,fHists[i]->GetMean(2));
      
      if (fHists[i]->GetEntries() < 10) {
         // low statistics
         fHists[i]->Clone();
         continue;         
      }

      // diagnosis before correction
      TH1* href = fHists[i]->ProjectionY();
      for (int ibinx = 0; ibinx < nBins; ibinx++) {
         TH1* hist = fHists[i]->ProjectionY("tmp",ibinx+1,ibinx+1);
         double distance = href->KolmogorovTest(hist,"M");
         hDiag->Fill(i,distance);
         delete hist ;
      }
      delete href;

      // rough correction using 1st bin
      scaled[i] = Util::ShiftCorrectionKT(fHists[i], scales, curvature);
      href = scaled[i]->ProjectionY();
      delete scaled[i];
      scaled[i] = Util::ShiftCorrectionKT(fHists[i], scales, curvature,href);
      // href is deleted by the function
      for (int j = 0, n = scales.size(); j < n ; ++j) {
         hEval->Fill(i,scales[j]);
      }
      href = scaled[i]->ProjectionY();
      // correction using total distribution
      for (int ibinx = 0; ibinx < nBins; ibinx++) {
         TH1* hist = scaled[i]->ProjectionY("tmp",ibinx+1,ibinx+1);
         double distance = href->KolmogorovTest(hist,"M");
         hCorrect->Fill(i,distance);
         delete hist;
      }
      delete href;
   }


   // diagnosis
   std::vector<bool>  isBad(nPads,false);
   double qave = 0.;
   int qnum = 0;
   
   for (int ipad = 0; ipad < nPads; ++ipad){
      double cnt = hHits->GetBinContent(ipad+1);
      if (cnt < 10.) {
         isBad[ipad] = true;
         continue;
      }
      qave += hMean->GetBinContent(ipad+1);
      qnum ++;
   }
   qave /= qnum;
   for (int ipad = 0; ipad < nPads; ++ipad){
      TCatReadoutPad *pad = static_cast<TCatReadoutPad*>(readoutPad->UncheckedAt(ipad));
      double sum = 0.;
      int nNeighbor = 0;
      double x1 = pad->X();
      double y1 = pad->Y();
      double cnt = hHits->GetBinContent(ipad+1);      
      for (int jpad = 0; jpad < nPads; ++jpad) {
         if (isBad[jpad] || ipad == jpad) continue;
         TCatReadoutPad *pad2 = static_cast<TCatReadoutPad*>(readoutPad->UncheckedAt(jpad));
         double dx = pad2->X() - x1;
         double dy = pad2->Y() - y1;
         if (dx * dx + dy + dy > 8 * 8) continue;
         sum += hHits->GetBinContent(jpad + 1);
         ++nNeighbor;
      }
      if (cnt < 0.5 * sum/nNeighbor || hMean->GetBinContent(ipad+1) < 0.1 * qave) isBad[ipad] = 1;
   }
   for (int ipad = 0; ipad < nPads; ++ipad) {
      if (isBad[ipad]) hMaybeBad->SetBinContent(ipad+1,1);
   }


   
   file->Write();
//   file->Close();
   Info(__func__,"Done");
}

